from flask import url_for

from kerko.shortcuts import composer


def build_sorter(criteria):
    sorter = {}
    active_sort_spec = criteria.get_active_sort_spec()
    if active_sort_spec:
        sorter['active'] = active_sort_spec.label
    sorter['items'] = [
        {
            'label':
                sort_spec.label,
            'url':
                url_for(
                    '.search', **criteria.params(options={
                        'sort': sort_spec.key,
                        'page': None,
                    })
                ),
        } for sort_spec in composer().get_ordered_specs('sorts')
        if sort_spec.is_allowed(criteria)
    ]

    return sorter



"""Update the search index from the local cache."""

import whoosh
from flask import current_app
from whoosh.query import Term

from kerko.shortcuts import composer, config
from kerko.storage import (SchemaError, SearchIndexError, load_object,
                           open_index, save_object)
from kerko.tags import TagGate


def sync_index():
    """
    Build the search index from the local cache.

    Return the number of synchronized items.
    """
    current_app.logger.info("Starting index sync...")
    library_context = load_object('cache', 'library')

    cache = open_index('cache')
    cache_version = load_object('cache', 'version', default=0)
    if not cache_version:
        raise SearchIndexError("The cache is empty and needs to be synchronized first.")
    # FIXME: The following does not detect when just the collections have changed in the cache (with no item changes). Should check the collections_version! https://pyzotero.readthedocs.io/en/latest/#zotero.Zotero.collection_versions
    # if load_object('index', 'version', default=0) == cache_version:
    #     current_app.logger.warning(f"The index is already up-to-date with cache version {cache_version}, nothing to do.")
    #     return 0

    def yield_items(parent_key):
        with cache.searcher() as searcher:
            results = searcher.search(Term('parentItem', parent_key), limit=None)
            for hit in results:
                yield hit.fields()

    def yield_top_level_items():
        return yield_items('')

    def yield_children(parent):
        return yield_items(parent['key'])

    count = 0
    index = open_index('index', schema=composer().schema, auto_create=True, write=True)
    writer = index.writer(limitmb=256)
    try:
        writer.mergetype = whoosh.writing.CLEAR
        gate = TagGate(
            config('kerko.zotero.item_include_re'),
            config('kerko.zotero.item_exclude_re'),
        )
        for item in yield_top_level_items():
            count += 1
            if gate.check(item['data']):
                item['children'] = list(yield_children(item))  # Extend the base Zotero item dict.
                document = {}
                for spec in list(composer().fields.values()) + list(composer().facets.values()):
                    spec.extract_to_document(document, item, library_context)
                writer.update_document(**document)
                current_app.logger.debug(
                    f"Item {count} updated ({item['key']}, {item.get('itemType')}): "
                    f"{document.get('data', {}).get('title')}"
                )
            else:
                current_app.logger.debug(f"Item {count} excluded ({item['key']})")
    except (whoosh.fields.FieldConfigurationError, whoosh.fields.UnknownFieldError) as e:
        writer.cancel()
        current_app.logger.error(e)
        raise SchemaError("Schema changes are required. Please clean index.") from e
    except Exception:  # pylint: disable=broad-except
        writer.cancel()
        raise
    else:
        writer.commit()
        # Save the cache's last_modified timestamp. Later, we cannot access the
        # cache directly to show the user when the data was last synchronized
        # from Zotero, because there is no guarantee that what the user's sees
        # (i.e., the current content of the index) is still in sync with the
        # cache (the cache might have been cleaned, or it might have been just
        # updated, with an index update still pending).
        save_object('index', 'last_update_from_zotero', cache.last_modified())
        save_object('index', 'version', cache_version)
        current_app.logger.info(
            f"Index sync successful, now at version {cache_version} "
            f"({count} top level item(s) processed)."
        )
    return count



import pprint
from datetime import datetime

import click
import wrapt
from flask import current_app
from flask.cli import with_appcontext

from kerko.storage import (SchemaError, SearchIndexError, delete_storage,
                           get_doc_count)
from kerko.sync import zotero
from kerko.sync.attachments import delete_attachments, sync_attachments
from kerko.sync.cache import sync_cache
from kerko.sync.index import sync_index


@wrapt.decorator
def execution_time_logger(wrapped, _instance, args, kwargs):
    start_time = datetime.now()
    return_value = wrapped(*args, **kwargs)
    current_app.logger.info(_format_elapsed_time(start_time))
    return return_value


@click.group()
def cli():
    """Run a Kerko subcommand."""


@cli.command()
@click.argument(
    'target',
    default='everything',
    type=click.Choice(['cache', 'index', 'attachments', 'everything'], case_sensitive=False),
)
@with_appcontext
@execution_time_logger
def sync(target):
    """
    Synchronize the cache, the search index, and/or the file attachments.

    By default, everything is synchronized.
    """
    try:
        if target in ['everything', 'cache']:
            sync_cache()
        if target in ['everything', 'index']:
            sync_index()
        if target in ['everything', 'attachments']:
            sync_attachments()
    except SearchIndexError as e:
        current_app.logger.error(e)
        raise click.Abort
    except SchemaError as e:
        current_app.logger.error(e)
        raise click.Abort


@cli.command()
@click.argument(
    'target',
    default='everything',
    type=click.Choice(['cache', 'index', 'attachments', 'everything'], case_sensitive=False),
)
@with_appcontext
def clean(target):
    """
    Delete the cache, the search index, and/or the attachments immediately.

    By default, everything is cleaned.
    """
    if target in ['everything', 'cache']:
        delete_storage('cache')
    if target in ['everything', 'index']:
        delete_storage('index')
    if target in ['everything', 'attachments']:
        delete_attachments()


@cli.command()
@click.argument(
    'target',
    type=click.Choice(['cache', 'index'], case_sensitive=False),
)
@with_appcontext
def count(target):
    """
    Show the number of records available in the cache or in the search index.

    The cache and the index are structured very differently and their respective
    numbers should not be expected to match. The number of records in the index
    may not even match the number of results obtained in Kerko's search
    interface, because the search can do some internal filtering.

    The cache is a flat database where items of any type or hierarchical level
    are counted as separate records.

    The index is a flat, denormalized database where each item is grouped with
    its children in a single record. The count may include items that are not
    usually displayed in search results, e.g., standalone notes or attachments.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    try:
        pprint.pprint(get_doc_count(target))
    except SearchIndexError as e:
        current_app.logger.error(e)
        raise click.Abort


@cli.command()
@with_appcontext
def config():
    """
    Show the app's full configuration.
    """
    pprint.pprint(dict(current_app.config))


@cli.command()
@click.argument('item_key')
@with_appcontext
def zotero_item(item_key):
    """
    Retrieve an item from the library, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item(credentials, item_key))


@cli.command()
@with_appcontext
def zotero_item_types():
    """
    List all item types, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_types(credentials))


@cli.command()
@with_appcontext
def zotero_item_fields():
    """
    List all fields, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_fields(credentials))


@cli.command()
@click.argument('item_type')
@with_appcontext
def zotero_item_type_fields(item_type):
    """
    List the available fields for a given item type, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_type_fields(credentials, item_type))


@cli.command()
@click.argument('item_type')
@with_appcontext
def zotero_item_type_creator_types(item_type):
    """
    List the available creator types for a given item type, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_type_creator_types(credentials, item_type))


@cli.command()
@with_appcontext
def zotero_top_level_collections():
    """
    List top-level collections of the library, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    collections = zotero.Collections(credentials, top_level=True)
    for c in collections:
        print(f"{c.get('key')} {c.get('data', {}).get('name', '')}")


def _format_elapsed_time(start_time):
    elapsed_time = int(round((datetime.now() - start_time).total_seconds()))
    elapsed_min, elapsed_sec = elapsed_time // 60, elapsed_time % 60
    s = 'Execution time:'
    if elapsed_min > 0:
        s += (' {n} minutes' if elapsed_min > 1 else ' {n} minute').format(n=elapsed_min)
        s += (' {n:02} seconds' if elapsed_sec > 1 else ' {n:02d} second').format(n=elapsed_sec)
    else:
        s += (' {n} seconds' if elapsed_sec > 1 else ' {n} second').format(n=elapsed_sec)
    return s



import math
import time
from collections import deque
from datetime import datetime, timedelta

from flask import (abort, current_app, flash, make_response, redirect,
                   render_template, request, send_from_directory, url_for)
from flask_babel import get_locale, gettext

from kerko import blueprint
from kerko.criteria import create_feed_criteria, create_search_criteria
from kerko.exceptions import except_abort
from kerko.forms import SearchForm
from kerko.searcher import Searcher
from kerko.shortcuts import composer, config
from kerko.specs import SortSpec
from kerko.storage import (SchemaError, SearchIndexError, get_doc_count,
                           get_storage_dir, load_object, open_index)
from kerko.views import pager
from kerko.views.item import build_item_context, creators, inject_item_data
from kerko.views.search import search_list, search_single

SITEMAP_URL_MAX_COUNT = 50000


@blueprint.route('/', methods=['GET', 'POST'])
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def search():
    """View the results of a search."""
    criteria = create_search_criteria(request.args)
    form = SearchForm(csrf_enabled=False)
    if form.validate_on_submit():
        scope = composer().scopes.get(form.scope.data)
        if scope:
            # Add the newly submitted keywords, and redirect to search with them.
            url = url_for(
                '.search',
                **criteria.params(
                    keywords=scope.add_keywords(form.keywords.data, criteria.keywords),
                    options={'page': None},
                )
            )
            return redirect(url, 302)

    if criteria.options.get('page-len', config('kerko.pagination.page_len')) == 1:
        return search_single(criteria, form)
    return search_list(criteria, form)


@blueprint.route('/atom.xml', methods=['GET'])
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def atom_feed():
    """Build a feed based on the search criteria."""

    if 'atom' not in config('kerko.feeds.formats'):
        return abort(404)

    context = {}
    criteria = create_feed_criteria(request.args)
    index = open_index('index')
    with Searcher(index) as searcher:
        extra_args = {}

        sort_field = composer().fields.get('sort_date_added')
        if sort_field:
            extra_args['sort_spec'] = SortSpec(
                key='date_added_desc',
                label='',
                fields=[sort_field],
                reverse=True,
            )
        else:
            current_app.logger.warning(
                "Feed cannot be sorted because the 'sort_date_added' was "
                "removed from the configuration."
            )

        if config('kerko.feeds.max_days'):
            if 'filter_date' in composer().fields:
                today = datetime.today()
                start = datetime(today.year, today.month,
                                 today.day) - timedelta(config('kerko.feeds.max_days'))
                current_app.logger.debug(
                    f"Show items dated {start} and newer"
                    f" (kerko.feeds.max_days == {config('kerko.feeds.max_days')})."
                )
                extra_args['require_date_ranges'] = {'filter_date': (start, None)}
            else:
                current_app.logger.warning(
                    "'kerko.feeds.max_days' is set but has no effect because the "
                    "'filter_date' field was removed from the configuration."
                )

        results = searcher.search_page(
            page=criteria.options.get('page', 1),
            page_len=criteria.options.get('page-len', config('kerko.pagination.page_len')),
            keywords=criteria.keywords,
            filters=criteria.filters,
            require_any=config('kerko.feeds.require_any'),  # Apply custom filtering, if not None.
            reject_any={'item_type': ['note', 'attachment'], **config('kerko.feeds.reject_any')},
            faceting=False,
            **extra_args,
        )
        if results.is_empty():
            items = []
        else:
            items = results.items(composer().select_fields(config('kerko.feeds.fields')))
            for item in items:
                creators.inject_creator_display_names(item)
        context['items'] = items
        context['total_count'] = results.item_count
        criteria.fit_page(results.page_count or 1)
        pager_sections = pager.get_sections(criteria.options['page'], results.page_count or 1)

    last_sync = load_object('index', 'last_update_from_zotero')
    if last_sync:
        context['last_sync'] = datetime.fromtimestamp(
            last_sync, tz=datetime.now().astimezone().tzinfo
        ).isoformat()
    else:
        context['last_sync'] = datetime.now().isoformat()

    if criteria.is_searching():
        context['feed_title'] = gettext('Custom feed')
    else:
        context['feed_title'] = gettext('Main feed')

    response = make_response(
        render_template(
            config('kerko.templates.atom_feed'),
            pager=pager.build_pager(
                pager_sections, criteria, endpoint='kerko.atom_feed', _external=True
            ),
            page_len=config('kerko.pagination.page_len'),
            feed_url=url_for(
                '.atom_feed',
                _external=True,
                **criteria.params(
                    options={
                        'page': criteria.options['page'] if criteria.options['page'] > 1 else None
                    }
                ),
            ),
            html_url=url_for(
                '.search',
                _external=True,
                **criteria.params(options={'page': None}),
            ),
            is_searching=criteria.is_searching(),
            locale=get_locale(),
            **context,
        )
    )
    response.headers['Content-Type'] = 'application/atom+xml; charset=utf-8'
    return response


@blueprint.route('/<path:item_id>')
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def item_view(item_id):
    """View a full bibliographic record."""
    start_time = time.process_time()

    index = open_index('index')
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(['id', 'alternate_id'])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={try_id_fields.popleft(): [item_id]},
                reject_any={'item_type': ['note', 'attachment']},
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().fields, composer().facets)[0]
        if fellback:
            return redirect(url_for('.item_view', item_id=item['id']), 301)
    inject_item_data(item)
    return render_template(
        config('kerko.templates.item'),
        **build_item_context(item),
        time=time.process_time() - start_time,
        locale=get_locale(),
    )


@blueprint.route('/<path:item_id>/download/<string:attachment_id>/')
@blueprint.route('/<path:item_id>/download/<string:attachment_id>/<string:attachment_filename>')
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def child_attachment_download(item_id, attachment_id, attachment_filename=None):
    """
    Download a child attachment.

    If the URL does not specify the attachment's filename or provides the wrong
    filename, a redirect is performed to a corrected URL so that the client gets
    a proper filename.
    """
    index = open_index('index')
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(['id', 'alternate_id'])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={try_id_fields.popleft(): [item_id]},
                reject_any={'item_type': ['note', 'attachment']},
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().select_fields(['id', 'attachments']))[0]
    matching_attachments = [a for a in item.get('attachments', []) if a.get('id') == attachment_id]
    if not matching_attachments or len(matching_attachments) > 1:
        flash(
            gettext(
                "The document you have requested has been removed. "
                "Please check below for the latest documents available."
            ),
            'warning',
        )
        return redirect(url_for('.item_view', item_id=item['id']), 301)
    attachment = matching_attachments[0]

    filepath = get_storage_dir('attachments') / attachment_id
    if not filepath.exists():
        return abort(404)

    filename = attachment['data'].get('filename', item['id'])
    if fellback or attachment_filename != filename:
        return redirect(url_for(
            '.child_attachment_download',
            item_id=item['id'],
            attachment_id=attachment_id,
            attachment_filename=filename,
        ), 301)

    return send_from_directory(
        get_storage_dir('attachments'),
        attachment_id,
        download_name=filename,
        mimetype=attachment['data'].get('contentType', 'octet-stream'),
    )


@blueprint.route('/download/<string:item_id>/')
@blueprint.route('/download/<string:item_id>/<string:attachment_filename>')
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def standalone_attachment_download(item_id, attachment_filename=None):
    """
    Download a standalone attachment.

    If the URL does not specify the attachment's filename or provides the wrong
    filename, a redirect is performed to a corrected URL so that the client gets
    a proper filename.
    """
    index = open_index('index')
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(['id', 'alternate_id'])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={
                    try_id_fields.popleft(): [item_id],
                    'item_type': ['attachment'],
                },
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().select_fields(['id', 'data']))[0]

    filepath = get_storage_dir('attachments') / item['id']
    if not filepath.exists():
        return abort(404)

    filename = item['data'].get('filename', item['id'])
    if fellback or attachment_filename != filename:
        return redirect(url_for(
            '.standalone_attachment_download',
            item_id=item['id'],
            attachment_filename=filename,
        ), 301)

    return send_from_directory(
        get_storage_dir('attachments'),
        item['id'],
        download_name=filename,
        mimetype=item['data'].get('contentType', 'octet-stream'),
    )


@blueprint.route('/<path:item_id>/export/<string:citation_format_key>')
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def item_citation_download(item_id, citation_format_key):
    """Download a record."""
    citation_format = composer().citation_formats.get(citation_format_key)
    if not citation_format:
        return abort(404)

    index = open_index('index')
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(['id', 'alternate_id'])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={try_id_fields.popleft(): [item_id]},
                reject_any={'item_type': ['note', 'attachment']},
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().select_fields(['id', citation_format.field.key]))[0]

    content = item.get(citation_format.field.key)
    if not content:
        return abort(404)

    if fellback:
        return redirect(
            url_for(
                '.item_citation_download',
                item_id=item['id'],
                citation_format_key=citation_format_key
            ), 301
        )

    response = make_response(content)
    response.headers['Content-Disposition'] = \
        f"attachment; filename={item['id']}.{citation_format.extension}"
    response.headers['Content-Type'] = \
        f'{citation_format.mime_type}; charset=utf-8'
    return response


@blueprint.route('/export/<string:citation_format_key>/')
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def search_citation_download(citation_format_key):
    """Download all records resulting from a search."""
    citation_format = composer().citation_formats.get(citation_format_key)
    if not citation_format:
        return abort(404)

    criteria = create_search_criteria(request.args)
    index = open_index('index')
    with Searcher(index) as searcher:
        results = searcher.search(
            limit=None,
            keywords=criteria.keywords,
            filters=criteria.filters,
            reject_any={'item_type': ['note', 'attachment']},
            sort_spec=criteria.get_active_sort_spec(),
            faceting=False,
        )
        if results.is_empty():
            return abort(404)
        citations = [
            item.get(citation_format.field.key, '') for item in
            results.items(field_specs={citation_format.field.key: citation_format.field})
        ]
    response = make_response(
        citation_format.group_format.format(
            citation_format.group_item_delimiter.join(citations)
        )
    )
    response.headers['Content-Disposition'] = \
        f'attachment; filename=bibliography.{citation_format.extension}'  # TODO: Make filename configurable.
    response.headers['Content-Type'] = \
        f'{citation_format.mime_type}; charset=utf-8'
    return response


def get_sitemap_page_count():
    count = get_doc_count('index')
    if count:
        count = math.ceil(count / SITEMAP_URL_MAX_COUNT)
        # Note: count may include items that will be excluded from the
        # sitemap, but this is the efficient way to get a good enough count.
    return count


@blueprint.route('/sitemap.xml')
@except_abort(SearchIndexError, 503)
def sitemap_index():
    """Generate a sitemap index."""
    response = make_response(
        render_template(
            'kerko/sitemap_index.xml.jinja2',
            sitemap_count=get_sitemap_page_count(),
        )
    )
    response.headers['Content-Type'] = 'application/xml; charset=utf-8'
    return response


@blueprint.route('/sitemap<int(min=1):page_num>.xml')
@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def sitemap(page_num):
    """Generate a sitemap."""
    if page_num > get_sitemap_page_count():
        return abort(404)

    index = open_index('index')
    with Searcher(index) as searcher:
        sort_field = composer().fields.get('sort_date_modified')
        if sort_field:
            sort_spec = SortSpec(
                key='date_added_desc',
                label='',
                fields=[sort_field],
                reverse=True,
            )
        else:
            sort_spec = None
        results = searcher.search_page(
            page=page_num,
            page_len=SITEMAP_URL_MAX_COUNT,
            reject_any={'item_type': ['note', 'attachment']},
            sort_spec=sort_spec,
            faceting=False,
        )
        if results.is_empty():
            return abort(404)
        items = results.items(composer().select_fields(['id', 'date_modified']))
    response = make_response(render_template('kerko/sitemap.xml.jinja2', items=items))
    response.headers['Content-Type'] = 'application/xml; charset=utf-8'
    return response


@blueprint.route('/api/last-sync')
def last_updated_on():
    last_sync = load_object('index', 'last_update_from_zotero')
    if last_sync:
        return {
            'when': datetime.fromtimestamp(
                last_sync, tz=datetime.now().astimezone().tzinfo
            ).isoformat(),
            'hours_ago': round((time.time() - last_sync) / 3600, 3),
        }
    return {}



"""
Unit tests for configuration helpers.
"""

import unittest

from flask import Config

from kerko import DEFAULTS as KERKO_DEFAULTS
from kerko.config_helpers import (KerkoModel, ConfigModel, config_get,
                                  config_set, config_update, parse_config)


class ConfigUpdateTestCase(unittest.TestCase):
    """Test config updating."""

    def setUp(self):
        self.config = Config(
            root_path='',
            defaults={'kerko': {'one': 1, 'two': 2}},
        )

    def test_update_defaults_no_overlap(self):
        config_update(self.config, {'kerko': {'ten': 10, 'twenty': 20, 'thirty': 30}})
        self.assertEqual(
            dict(self.config),
            {'kerko': {'one': 1, 'two': 2, 'ten': 10, 'twenty': 20, 'thirty': 30}}
        )

    def test_update_defaults_with_overlap(self):
        config_update(self.config, {'kerko': {'two': 20, 'thirty': 30}})
        self.assertEqual(dict(self.config), {'kerko': {'one': 1, 'two': 20, 'thirty': 30}})


class ConfigGetSetTestCase(unittest.TestCase):
    """Test config access."""

    def setUp(self):
        self.config = Config(
            root_path='',
            defaults={'kerko': {'one': 1, 'none': None, 'list': ['a', 'b', 'c']}},
        )

    def test_get_existing(self):
        self.assertEqual(config_get(self.config, 'kerko.one'), 1)
        self.assertEqual(config_get(self.config, 'kerko.none'), None)
        self.assertEqual(config_get(self.config, 'kerko.list'), ['a', 'b', 'c'])

    def test_get_nonexisting(self):
        with self.assertRaises(KeyError):
            config_get(self.config, 'kerko.foo')

    def test_set_new_value(self):
        config_set(self.config, 'kerko.foo', 'bar')
        self.assertEqual(
            dict(self.config),
            {'kerko': {'one': 1, 'none': None, 'list': ['a', 'b', 'c'], 'foo': 'bar'}},
        )

    def test_set_replace_value(self):
        config_set(self.config, 'kerko.one', 10)
        self.assertEqual(
            dict(self.config),
            {'kerko': {'one': 10, 'none': None, 'list': ['a', 'b', 'c']}},
        )

    def test_set_replace_list(self):
        config_set(self.config, 'kerko.list', ['x', 'y', 'z'])
        self.assertEqual(
            dict(self.config),
            {'kerko': {'one': 1, 'none': None, 'list': ['x', 'y', 'z']}},
        )


class ParseKerkoConfigTestCase(unittest.TestCase):
    """Test parsing of the kerko config table."""

    def test_default_config(self):
        config = Config(root_path='', defaults=KERKO_DEFAULTS)
        parse_config(config, 'kerko', KerkoModel)

    def test_invalid_config(self):
        config = Config(root_path='', defaults={'kerko': 'bad_config'})
        with self.assertRaises(RuntimeError):
            parse_config(config, 'kerko', KerkoModel)


class ParseRootConfigTestCase(unittest.TestCase):
    """Test parsing of the config root."""

    def test_valid_config(self):
        config = Config(
            root_path='',
            defaults={
                'SECRET_KEY': '1234567890AB',
                'ZOTERO_API_KEY': '1234567890ABCDEFGHIJKLMN',
                'ZOTERO_LIBRARY_ID': '123456',
                'ZOTERO_LIBRARY_TYPE': 'group',
                'ignored1': '1',
                'ignored2': 2,
                'ignored3': [1, 2, 'three'],
            }
        )
        parse_config(config, None, ConfigModel)
        self.assertEqual(config['ignored1'], '1')
        self.assertEqual(config['ignored2'], 2)
        self.assertEqual(config['ignored3'], [1, 2, 'three'])

    def test_coerced_variable_type(self):
        config = Config(
            root_path='',
            defaults={
                'SECRET_KEY': '1234567890AB',
                'ZOTERO_API_KEY': '1234567890ABCDEFGHIJKLMN',
                'ZOTERO_LIBRARY_ID': 123456,  # Integer.
                'ZOTERO_LIBRARY_TYPE': 'group',
            }
        )
        parse_config(config, None, ConfigModel)
        self.assertEqual(config['ZOTERO_LIBRARY_ID'], '123456')

    def test_missing_variable(self):
        config = Config(
            root_path='',
            defaults={
                'SECRET_KEY': '1234567890AB',
                'ZOTERO_LIBRARY_ID': '123456',
                'ZOTERO_LIBRARY_TYPE': 'group',
                # Missing 'ZOTERO_API_KEY'.
            }
        )
        with self.assertRaises(RuntimeError):
            parse_config(config, None, ConfigModel)

    def test_empty_variable(self):
        config = Config(
            root_path='',
            defaults={
                'SECRET_KEY': '',  # Empty.
                'ZOTERO_API_KEY': '1234567890ABCDEFGHIJKLMN',
                'ZOTERO_LIBRARY_ID': '123456',
                'ZOTERO_LIBRARY_TYPE': 'group',
            }
        )
        with self.assertRaises(RuntimeError):
            parse_config(config, None, ConfigModel)

    def test_invalid_variable(self):
        config = Config(
            root_path='',
            defaults={
                'SECRET_KEY': '1234567890AB',
                'ZOTERO_API_KEY': '1234567890ABCDEFGHIJKLMN',
                'ZOTERO_LIBRARY_ID': 'ABCDEF',  # Invalid.
                'ZOTERO_LIBRARY_TYPE': 'group',
            }
        )
        with self.assertRaises(RuntimeError):
            parse_config(config, None, ConfigModel)



import abc
import pathlib
from typing import Any, Dict, List, Optional, Type, Union

from typing_extensions import Annotated, Literal

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

import dpath
import whoosh
from flask import Config
from pydantic import (BaseModel,  # pylint: disable=no-name-in-module
                      ConstrainedStr, Extra, Field, NonNegativeInt,
                      ValidationError, validator)

# pylint: disable=too-few-public-methods

# Note: To preserve field ordering in Pydantic models, we annotate an
# attribute's type even when it could be determined by its default value.
# See https://docs.pydantic.dev/latest/usage/models/#field-ordering


class SlugStr(ConstrainedStr):

    regex = r'^[a-z][a-z0-9_\-]*$'


class FieldNameStr(ConstrainedStr):

    regex = r'^[a-z][a-zA-Z0-9_]*$'


class ElementIdStr(ConstrainedStr):

    regex = r'^[a-z][a-zA-Z0-9]*$'


class AssetsModel(BaseModel):
    """Model for the kerko.assets config table."""

    class Config:
        extra = Extra.forbid

    bootstrap_version: str
    jquery_version: str
    popper_version: str
    with_jquery: bool
    with_popper: bool


class FeaturesModel(BaseModel):
    """Model for the kerko.features config table."""

    class Config:
        extra = Extra.forbid

    download_attachment_new_window: bool
    download_citations_link: bool
    download_citations_max_count: bool
    open_in_zotero_app: bool
    open_in_zotero_web: bool
    print_citations_link: bool
    print_citations_max_count: bool
    print_item_link: bool
    relations_links: bool
    relations_initial_limit: int = Field(ge=5)
    relations_sort: SlugStr
    results_abstracts: bool
    results_abstracts_max_length: NonNegativeInt
    results_abstracts_max_length_leeway: NonNegativeInt
    results_abstracts_toggler: bool
    results_attachment_links: bool
    results_url_links: bool


class FeedsModel(BaseModel):
    """Model for the kerko.feeds config table."""

    class Config:
        extra = Extra.forbid

    formats: List[Optional[Literal["atom"]]]
    fields: List[FieldNameStr]
    max_days: NonNegativeInt
    require_any: Dict[FieldNameStr, List[Union[str, bool, int, float]]]
    reject_any: Dict[FieldNameStr, List[Union[str, bool, int, float]]]


class MetaModel(BaseModel):
    """Model for the kerko.meta config table."""

    class Config:
        extra = Extra.forbid

    title: str
    highwirepress_tags: bool
    google_analytics_id: str


class PaginationModel(BaseModel):
    """Model for the kerko.pagination config table."""

    class Config:
        extra = Extra.forbid

    page_len: int = Field(ge=2)
    pager_links: int = Field(ge=2)


class TemplatesModel(BaseModel):
    """Model for the kerko.templates config table."""

    class Config:
        extra = Extra.forbid

    base: str
    layout: str
    search: str
    search_item: str
    item: str
    atom_feed: str


class ZoteroModel(BaseModel):
    """Model for the kerko.zotero config table."""

    class Config:
        extra = Extra.forbid

    batch_size: int = Field(ge=20)
    max_attempts: int = Field(ge=1)
    wait: int = Field(ge=120)
    csl_style: SlugStr
    locale: str = Field(regex=r'^[a-z]{2}-[A-Z]{2}$')
    item_include_re: str
    item_exclude_re: str
    tag_include_re: str
    tag_exclude_re: str
    child_include_re: str
    child_exclude_re: str
    attachment_mime_types: List[str]


class SearchModel(BaseModel):
    """Model for the kerko.search config table."""

    class Config:
        extra = Extra.forbid

    result_fields: List[FieldNameStr]
    fulltext: bool
    whoosh_language: str = Field(regex=r'^[a-z]{2,3}$')

    @validator('whoosh_language')
    def validate_whoosh_has_language(cls, v):  # pylint: disable=no-self-argument
        if not whoosh.lang.has_stemmer(v):
            raise ValueError(f"language '{v}' not supported by Whoosh")
        return v


class ScopesModel(BaseModel):
    """Model for the kerko.scopes config table."""

    class Config:
        extra = Extra.forbid

    enabled: bool = True
    selector_label: Optional[str]
    breadbox_label: Optional[str]
    help_text: Optional[str]
    weight: int = 0


class CoreRequiredSearchFieldModel(BaseModel):
    """Model for the kerko.search_fields.core.required config table."""

    class Config:
        extra = Extra.forbid

    scopes: List[SlugStr]
    boost: float


class CoreOptionalSearchFieldModel(BaseModel):
    """Model for the kerko.search_fields.core.optional config table."""

    class Config:
        extra = Extra.forbid

    enabled: bool = True
    scopes: List[SlugStr]
    boost: float


class ZoteroFieldModel(BaseModel):
    """Model for the kerko.search_fields.zotero config table."""

    class Config:
        extra = Extra.forbid

    enabled: bool = True
    scopes: List[SlugStr]
    analyzer: Union[Literal["id"], Literal["text"], Literal["name"]]
    boost: float


class CoreSearchFieldsModel(BaseModel):

    class Config:
        extra = Extra.forbid

    required: Dict[FieldNameStr, CoreRequiredSearchFieldModel]
    optional: Dict[FieldNameStr, CoreOptionalSearchFieldModel]


class SearchFieldsModel(BaseModel):
    """Base model for the kerko.search_fields config table."""

    class Config:
        extra = Extra.forbid

    core: CoreSearchFieldsModel
    zotero: Dict[FieldNameStr, ZoteroFieldModel]


class BaseFacetModel(BaseModel, abc.ABC):
    """Base model for the kerko.facets config table."""

    class Config:
        extra = Extra.forbid

    enabled: bool = True
    filter_key: SlugStr
    weight: int = 0
    initial_limit: NonNegativeInt = 0
    initial_limit_leeway: NonNegativeInt = 2
    sort_by: List[Union[Literal["label"], Literal["count"]]] = ["label"]
    sort_reverse: bool = False
    item_view: bool = True


class TagFacetModel(BaseFacetModel):

    type: Literal["tag"]
    title: Optional[str]


class ItemTypeFacetModel(BaseFacetModel):
    type: Literal["item_type"]
    title: Optional[str]
    item_view: bool = False


class YearFacetModel(BaseFacetModel):
    type: Literal["year"]
    title: Optional[str]
    item_view: bool = False


class LinkFacetModel(BaseFacetModel):
    type: Literal["link"]
    title: Optional[str]
    item_view: bool = False


class CollectionFacetModel(BaseFacetModel):
    type: Literal["collection"]
    title: str
    collection_key: str = Field(regex=r'^[A-Z]{8}$')


# Note: Discriminated unions ensure that a single unambiguous error gets
# reported when validation fails. Reference:
# https://docs.pydantic.dev/latest/usage/types/#discriminated-unions-aka-tagged-unions

FacetModelUnion = Annotated[
    Union[
        TagFacetModel,
        ItemTypeFacetModel,
        YearFacetModel,
        LinkFacetModel,
        CollectionFacetModel,
    ], Field(discriminator='type')
]


class SortsModel(BaseModel):
    """Model for the kerko.sorts config table."""

    class Config:
        extra = Extra.forbid

    enabled: bool = True
    weight: int = 0
    label: Optional[str]


class CitationFormatsModel(BaseModel):
    """Model for the kerko.citation_formats config table."""

    class Config:
        extra = Extra.forbid

    enabled: bool = True
    weight: int = 0
    label: Optional[str]
    help_text: Optional[str]
    extension: SlugStr
    mime_type: str


class RelationsModel(BaseModel):
    """Model for the kerko.relations config table."""

    class Config:
        extra = Extra.forbid

    enabled: bool = True
    weight: int = 0
    label: Optional[str]


class KerkoModel(BaseModel):
    """Model for the kerko config table."""

    class Config:
        extra = Extra.forbid

    assets: AssetsModel
    features: FeaturesModel
    feeds: FeedsModel
    meta: MetaModel
    pagination: PaginationModel
    templates: TemplatesModel
    zotero: ZoteroModel
    search: SearchModel
    scopes: Dict[SlugStr, ScopesModel]
    search_fields: SearchFieldsModel
    facets: Dict[FieldNameStr, FacetModelUnion]
    sorts: Dict[SlugStr, SortsModel]
    citation_formats: Dict[SlugStr, CitationFormatsModel]
    relations: Dict[ElementIdStr, RelationsModel]


class ConfigModel(BaseModel):
    """Model for validating mandatory at the root config table."""

    # Note: This model allows extra fields since we cannot cover all variables
    # that Flask, Flask extensions, and applications might support.

    SECRET_KEY: str = Field(min_length=12)
    ZOTERO_API_KEY: str = Field(min_length=16)
    ZOTERO_LIBRARY_ID: str = Field(regex=r'^[0-9]+$')
    ZOTERO_LIBRARY_TYPE: Union[Literal["user"], Literal["group"]]
    kerko: Optional[KerkoModel]


def load_toml(filename: Union[str, pathlib.Path], verbose=False) -> Dict[str, Any]:
    """Load the content of a TOML file."""
    try:
        with open(filename, 'rb') as file:
            config = tomllib.load(file)
            if verbose:
                print(f" * Loading configuration file {filename}")
            return config
    except OSError as e:
        raise RuntimeError(f"Unable to open TOML file.\n{e}") from e
    except tomllib.TOMLDecodeError as e:
        raise RuntimeError(f"Invalid TOML format in file '{filename}'.\n{e}") from e


def config_update(config: Config, new_data: Dict[str, Any]) -> None:
    """
    Update the configuration with the specified `dict`.

    Unlike the standard `dict.update`, this performs a deep merge. However, it
    does not deep copy the objects.
    """
    dpath.merge(config, new_data, flags=dpath.MergeType.REPLACE)


def config_get(config: Config, path: str) -> Any:
    """
    Retrieve an arbitrarily nested configuration parameter.

    `path` is a string of keys separated by dots ('.') acting as hierarchical
    level separators.

    Unlike `dict.get()`, this provides no way of specifying a default value as
    all configuration parameters are expected to have be already initialized
    with a default value. Passing an invalid `path`, i.e., trying to access a
    parameter that does not have a value, is considered a programming error and
    will throw a `KeyError` exception.
    """
    return dpath.get(config, path, separator='.')


def config_set(config: Config, path: str, value: Any) -> None:
    """
    Set an arbitrarily nested configuration parameter.

    `path` is a string of keys separated by dots ('.') acting as hierarchical
    level separators.
    """
    dpath.new(config, path, value, separator='.')


def parse_config(
    config: Config,
    key: Optional[str] = None,
    model: Type[BaseModel] = ConfigModel
) -> None:
    """
    Parse and validate configuration using `model`.

    Parameter values get replaced by parsed ones. Values may thus get silently
    coerced into the types specified by the model (unless strict typing is
    enforced by the model, in which case an error will be raised).

    If `key` is `None`, then the whole configuration gets parsed with the given
    `model`. Otherwise only the structure at the specified `key` gets parsed.

    If `key` does not exists in the config, parsing is silently skipped.
    """
    try:
        if key is None:
            config.update(model.parse_obj(config).dict())
        elif config.get(key):
            # The parsed models are stored in the config as dicts. This way, the
            # whole configuration structure is made of dicts only, allowing
            # consistent access for any element at any depth.
            config_set(config, key, model.parse_obj(config[key]).dict())
    except ValidationError as e:
        raise RuntimeError(f"Invalid configuration. {e}") from e



import time
from datetime import datetime

from babel.numbers import format_decimal
from flask import redirect, render_template, url_for
from flask_babel import get_locale, gettext, ngettext
from werkzeug.datastructures import MultiDict

from kerko.criteria import create_feed_criteria
from kerko.searcher import Searcher
from kerko.shortcuts import composer, config
from kerko.storage import load_object, open_index
from kerko.views import breadbox, pager, sorter
from kerko.views.item import build_item_context, inject_item_data


def _build_item_search_urls(items, criteria):
    """Generate an URL for each search result (switching the search to page-len=1)."""
    def build_url(item, position):
        return url_for(
            '.search',
            **criteria.params(
                options={
                    'page': (criteria.options.get('page', 1) - 1) * page_len + position + 1,
                    'page-len': 1,
                    'id': item['id'],
                }
            )
        )

    page_len = criteria.options.get('page-len', config('kerko.pagination.page_len'))
    if page_len == 'all':
        page_len = 0
    return [build_url(item, position) for position, item in enumerate(items)]


def _get_page_item_ids(criteria, page_numbers, current_item_id):
    """Retrieve the id of each item corresponding to the specified search result pages."""
    assert criteria.options.get('page-len') == 1
    item_ids = {}  # Dict of item ids, keyed by page number.
    index = open_index('index')
    with Searcher(index) as searcher:
        for p in page_numbers:
            if p == criteria.options.get('page', 1):
                # We already know the current page's item id. No further query necessary.
                item_ids[p] = current_item_id
            else:
                # Run a search query to get the item id corresponding to the page number.
                page_results = searcher.search_page(
                    page=p,
                    page_len=1,
                    keywords=criteria.keywords,
                    filters=criteria.filters,
                    reject_any={'item_type': ['note', 'attachment']},
                    sort_spec=criteria.get_active_sort_spec(),
                    faceting=False,
                )
                item_ids[p] = page_results.items(composer().select_fields(['id']))[0]['id']
    return item_ids


def empty_results(criteria, form):
    """Prepare the template context variables for an empty search results page."""
    context = {}
    facets = {}
    context['title'] = gettext('Your search did not match any resources')
    if criteria.has_filters():
        # When search results are empty, Whoosh cannot provide any groupings,
        # thus our facets cannot be built. However, we still want to display the
        # breadbox with active filters, if any. To build that, we perform a
        # separate search query for each active facet, each time ignoring all
        # other search criteria. Unless a given facet value alone leads to empty
        # results, we'll be able to get to build that facet.
        index = open_index('index')
        with Searcher(index) as searcher:
            for key, value in criteria.filters.lists():
                results = searcher.search(
                    filters=MultiDict({key: value}),
                    reject_any={'item_type': ['note', 'attachment']},
                    limit=1,  # We don't care about the actual items.
                    faceting=True,
                )
                facets.update(results.facets(composer().facets, criteria, active_only=True))
    context['breadbox'] = breadbox.build_breadbox(criteria, facets)
    return render_template(
        config('kerko.templates.search'),
        form=form,
        locale=get_locale(),
        is_searching=criteria.is_searching(),
        **context,
    )


def search_single(criteria, form):
    """Perform search, and prepare template context for a results page containing a single item."""
    start_time = time.process_time()
    context = {}
    index = open_index('index')
    with Searcher(index) as searcher:
        results = searcher.search_page(
            page=criteria.options.get('page', 1),
            page_len=1,
            keywords=criteria.keywords,
            filters=criteria.filters,
            reject_any={'item_type': ['note', 'attachment']},
            sort_spec=criteria.get_active_sort_spec(),
            faceting=True,
        )

        criteria_id = criteria.options.get('id')
        if criteria_id and (results.is_empty() or criteria_id != results[0]['id']):
            # The search URL is obsolete, the result no longer matches the expected item.
            return redirect(url_for('.item_view', item_id=criteria_id, _external=True), 301)

        if results.is_empty():
            return empty_results(criteria, form)

        criteria.fit_page(results.page_count)
        if criteria.is_searching():
            context['search_title'] = gettext('Your search')
        else:
            context['search_title'] = gettext('Full bibliography')

        # Load item with all available fields.
        item = results.items(composer().fields, composer().facets)[0]
        facets = results.facets(composer().facets, criteria, active_only=True)

        inject_item_data(item)
        context.update(build_item_context(item))
        context['total_count'] = results.item_count
        context['total_count_formatted'] = format_decimal(results.item_count, locale=get_locale())
        context['page_count'] = results.page_count
        context['page_count_formatted'] = format_decimal(results.page_count, locale=get_locale())
        pager_sections = pager.get_sections(criteria.options['page'], results.page_count)
        page_item_ids = _get_page_item_ids(
            criteria, pager.get_page_numbers(pager_sections), item['id']
        )
        context['pager'] = pager.build_pager(
            pager_sections,
            criteria,
            page_options={p: {'id': item_id} for p, item_id in page_item_ids.items()},
        )
        context['breadbox'] = breadbox.build_breadbox(criteria, facets)
        context['back_url'] = url_for(
            '.search',
            **criteria.params(
                options={
                    'page': int(
                        (criteria.options.get('page', 1) - 1) / config('kerko.pagination.page_len') + 1
                    ),
                    'page-len': None,
                    'id': None,
                }
            )
        )
    return render_template(
        config('kerko.templates.search_item'),
        form=form,
        time=time.process_time() - start_time,
        locale=get_locale(),
        is_searching=criteria.is_searching(),
        **context,
    )


def search_list(criteria, form):
    """Perform search, and prepare the template context variables for a list of search results."""
    start_time = time.process_time()
    context = {}
    index = open_index('index')
    with Searcher(index) as searcher:
        page_len = criteria.options.get('page-len', config('kerko.pagination.page_len'))
        common_search_args = {
            'keywords': criteria.keywords,
            'filters': criteria.filters,
            'reject_any': {'item_type': ['note', 'attachment']},
            'sort_spec': criteria.get_active_sort_spec(),
            'faceting': True,
        }
        if page_len == 'all':
            results = searcher.search(
                limit=None,
                **common_search_args,
            )
        else:
            results = searcher.search_page(
                page=criteria.options.get('page', 1),
                page_len=page_len,
                **common_search_args,
            )

        if results.is_empty():
            return empty_results(criteria, form)

        criteria.fit_page(results.page_count)

        if criteria.is_searching():
            context['title'] = ngettext('Result', 'Results', results.item_count)
        else:
            context['title'] = gettext('Full bibliography')
        context['total_count'] = results.item_count
        context['total_count_formatted'] = format_decimal(results.item_count, locale=get_locale())
        context['page_count'] = results.page_count
        context['page_count_formatted'] = format_decimal(results.page_count, locale=get_locale())
        context['pager'] = pager.build_pager(
            pager.get_sections(criteria.options['page'], results.page_count),
            criteria,
        )
        context['sorter'] = sorter.build_sorter(criteria)
        context['show_abstracts'] = criteria.options.get(
            'abstracts', config('kerko.features.results_abstracts')
        )
        context['abstracts_toggler_url'] = url_for(
            '.search',
            **criteria.params(options={
                'abstracts': int(
                    not criteria.options.
                    get('abstracts', config('kerko.features.results_abstracts'))
                ),
            })
        )
        context['print_url'] = url_for(
            '.search',
            **criteria.params(options={
                'page': None,
                'page-len': 'all',
                'print-preview': 1,
            })
        )
        context['print_preview'] = criteria.options.get('print-preview', 0)
        context['download_urls'] = {
            key: url_for(
                '.search_citation_download',
                citation_format_key=key,
                **criteria.params(options={
                    'page': None,
                })
            )
            for key in composer().citation_formats.keys()
        }
        if 'atom' in config('kerko.feeds.formats'):
            context['atom_feed_url'] = url_for(
                '.atom_feed',
                **create_feed_criteria(initial=criteria).params(options={
                    'page': None,
                })
            )
            if criteria.is_searching():
                context['atom_feed_help'] = gettext('Custom feed based on your search')
                context['atom_feed_title'] = gettext('Custom feed')
            else:
                context['atom_feed_title'] = gettext('Main feed')

        # Prepare search result items.
        field_specs = composer().select_fields(
            config('kerko.search.result_fields') +
            [badge.field.key for badge in composer().badges.values()],
        )
        items = results.items(field_specs)
        results_facets = results.facets(composer().facets, criteria)
        context['search_results'] = zip(items, _build_item_search_urls(items, criteria))
        context['facet_results'] = results_facets
        context['breadbox'] = breadbox.build_breadbox(criteria, results_facets)

        last_sync = load_object('index', 'last_update_from_zotero')
        if last_sync:
            context['last_sync'] = datetime.fromtimestamp(
                last_sync,
                tz=datetime.now().astimezone().tzinfo,
            )
    return render_template(
        config('kerko.templates.search'),
        form=form,
        time=time.process_time() - start_time,
        locale=get_locale(),
        is_searching=criteria.is_searching(),
        **context,
    )



"""
Tests for item meta tags.
"""

import unittest

import elementpath  # For XPath 2.0 selectors.
from lxml import etree
from kerko.views.item.creators import format_creator_name

from tests.integration_testing import SynchronizedTestCase


class CreatorNameTestCase(unittest.TestCase):
    """Unit test for creator name formatting."""

    def test_creator_empty(self):
        self.assertEqual(
            format_creator_name({}),
            '',
        )

    def test_creator_single_field(self):
        self.assertEqual(
            format_creator_name({'name': 'Full name'}),
            'Full name',
        )

    def test_creator_single_field_empty(self):
        self.assertEqual(
            format_creator_name({'name': ''}),
            '',
        )

    def test_creator_two_fields(self):
        self.assertEqual(
            format_creator_name({'firstName': 'FirstName', 'lastName': 'LastName'}),
            'LastName, FirstName',
        )

    def test_creator_two_fields_empty(self):
        self.assertEqual(
            format_creator_name({'firstName': '', 'lastName': ''}),
            '',
        )


class ItemMetaTestCase(SynchronizedTestCase):
    """Integration tests for item meta tags."""

    def test_highwire_press_book(self):
        """Test Highwire Press tags for a 'book' item."""

        with self.app.test_client() as client:
            response = client.get(f'{self.URL_PREFIX}/2351708-TAQ6HCSL', follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ['A book with multiple authors'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ['2019'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ['2019'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ['2019'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                [
                    'Mustermann, Erika',
                    'Jacques, Pierre-Jean',
                    'Pérez, Juan',
                    'Doe, John',
                    'Roe, Richard',
                    'Smith, Jane',
                    'Kumar, Ashok',
                ],
                # Google Scholar requires that only the actual authors be included.
                # Thus, 'Jansen, Jan' (Editor) is excluded.
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ['en'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publisher"]/@content'),
                ['Example Publishing'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_isbn"]/@content'),
                ['979-8292-9-6571-8'],
            )
            unexpected_tags = [
                'citation_doi',
                'citation_issn',
                'citation_volume',
                'citation_issue',
                'citation_pdf_url',
                'citation_firstpage',
                'citation_lastpage',
                'citation_conference_title',
                'citation_journal_title',
                'citation_dissertation_institution',
                'citation_technical_report_institution',
                'citation_technical_report_number',
            ]
            self.assertEqual(
                elementpath.select(
                    tree,
                    f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                "Tag is unexpected for this particular item, or for 'book' items in general"
            )

    def test_highwire_press_conference_paper(self):
        """Test Highwire Press tags for a 'conferencePaper' item."""

        with self.app.test_client() as client:
            response = client.get(f'{self.URL_PREFIX}/2351708-99VVIN8V', follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ['A conference paper'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ['2021-11-10'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ['2021-11-10'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ['2021'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ['Hoe, Harry'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ['en'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publisher"]/@content'),
                ['Example Publishing'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_firstpage"]/@content'),
                ['11'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_lastpage"]/@content'),
                ['55'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_conference_title"]/@content'),
                ['First Example Conference, 2021'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_doi"]/@content'),
                ['10.5555/2351708-99VVIN8V'],
            )
            unexpected_tags = [
                'citation_isbn',
                'citation_issn',
                'citation_volume',
                'citation_issue',
                'citation_pdf_url',
                'citation_journal_title',
                'citation_dissertation_institution',
                'citation_technical_report_institution',
                'citation_technical_report_number',
                "Tag is unexpected for this particular item, "
                "or for 'conferencePaper' items in general"
            ]
            self.assertEqual(
                elementpath.select(
                    tree,
                    f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
            )

    def test_highwire_press_journal_article(self):
        """Test Highwire Press tags for a 'journalArticle' item."""

        with self.app.test_client() as client:
            response = client.get(f'{self.URL_PREFIX}/2351708-NW89E453', follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ['Zotero: information management software 2.0'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ['2011-06-07'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ['2011-06-07'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ['2011'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ['Fernandez, Peter'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_volume"]/@content'),
                ['28'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_issue"]/@content'),
                ['4'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ['en'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_firstpage"]/@content'),
                ['5'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_lastpage"]/@content'),
                ['7'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_journal_title"]/@content'),
                ['Library Hi Tech News'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_doi"]/@content'),
                ['10.1108/07419051111154758'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_issn"]/@content'),
                ['0741-9058'],
            )
            self.assertTrue(
                elementpath.select(
                    tree,
                    f'/html/head/meta[@name="citation_pdf_url"][matches(@content, '
                    f'"https?://[^/]+{self.URL_PREFIX}/[A-Z0-9]+/download/[A-Z0-9]+/Fernandez'
                    f'%20-%202011%20-%20Zotero%20information%20management%20software%202.0.pdf")]'
                ),
                "Incorrect PDF URL"
            )
            unexpected_tags = [
                'citation_isbn',
                'citation_publisher',
                'citation_conference_title',
                'citation_dissertation_institution',
                'citation_technical_report_institution',
                'citation_technical_report_number',
            ]
            self.assertEqual(
                elementpath.select(
                    tree,
                    f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                "Tag is unexpected for this particular item, or for 'journalArticle' items in general"
            )

    def test_highwire_press_report(self):
        """Test Highwire Press tags for a 'report' item."""

        with self.app.test_client() as client:
            response = client.get(f'{self.URL_PREFIX}/2351708-DNFRSHFB', follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ['A report with a DOI in the Extra field'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ['2019'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ['2019'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ['2019'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ['Doe, John'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ['en'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_technical_report_institution"]/@content'),
                ['Example Institution'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_technical_report_number"]/@content'),
                ['10'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_doi"]/@content'),
                ['10.5555/2351708-DNFRSHFB'],
            )
            unexpected_tags = [
                'citation_isbn',
                'citation_issn',
                'citation_pdf_url',
                'citation_volume',
                'citation_issue',
                'citation_firstpage',
                'citation_lastpage',
                'citation_publisher',
                'citation_journal_title',
                'citation_conference_title',
                'citation_dissertation_institution',
            ]
            self.assertEqual(
                elementpath.select(
                    tree,
                    f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                "Tag is unexpected for this particular item, or for 'report' items in general"
            )

    def test_highwire_press_thesis(self):
        """Test Highwire Press tags for a 'thesis' item."""

        with self.app.test_client() as client:
            response = client.get(f'{self.URL_PREFIX}/2351708-9TCEM5BE', follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_title"]/@content'),
                ['A thesis example'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_publication_date"]/@content'),
                ['2015-05-15'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_date"]/@content'),
                ['2015-05-15'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_year"]/@content'),
                ['2015'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_author"]/@content'),
                ['Hoe, Harry'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_language"]/@content'),
                ['en'],
            )
            self.assertEqual(
                tree.xpath('/html/head/meta[@name="citation_dissertation_institution"]/@content'),
                ['Small University'],
            )
            unexpected_tags = [
                'citation_doi',
                'citation_isbn',
                'citation_issn',
                'citation_pdf_url',
                'citation_volume',
                'citation_issue',
                'citation_firstpage',
                'citation_lastpage',
                'citation_publisher',
                'citation_journal_title',
                'citation_conference_title',
                'citation_technical_report_institution',
                'citation_technical_report_number',
            ]
            self.assertEqual(
                elementpath.select(
                    tree,
                    f'/html/head/meta[matches(@name, "{"|".join(unexpected_tags)}")]/@name'
                ),
                [],
                "Tag is unexpected for this particular item, or for 'thesis' items in general"
            )

    def test_highwire_press_webpage(self):
        """Test Highwire Press tags for a 'webpage' item."""

        with self.app.test_client() as client:
            response = client.get(f'{self.URL_PREFIX}/4507457-YE4WVE35', follow_redirects=True)
            self.assertEqual(response.status_code, 200)

            parser = etree.HTMLParser()
            tree = etree.fromstring(response.get_data(as_text=True), parser)
            self.assertEqual(
                elementpath.select(
                    tree,
                    '/html/head/meta[matches(@name, "citation_.+")]/@name'
                ),
                [],
                "Only book, conferencePaper, journalArticle, report, or thesis "
                "items are allowed to have Highwire Press tags"
            )

    # TODO: Test item with empty title/author/date.



"""
Kerko: A Flask blueprint that provides faceted search for bibliographies based on Zotero.
"""
# pylint: disable=invalid-name

import pathlib
import sys

from flask import Blueprint

from kerko.config_helpers import load_toml
from kerko.jinja2 import register_filters

# Kerko won't load translations on its own. To load them, an application may add
# the following domain and translation directories to its Babel configuration.
TRANSLATION_DOMAIN = 'kerko'
TRANSLATION_DIRECTORIES = [str(pathlib.Path(__file__).parent / 'translations')]

try:
    DEFAULTS = load_toml(pathlib.Path(__file__).parent / 'default_config.toml')
except RuntimeError as e:
    print(e, file=sys.stderr)
    sys.exit(1)


def make_blueprint():
    b = Blueprint(
        'kerko', __name__, static_folder='static', template_folder='templates'
    )
    register_filters(b)
    return b


blueprint = make_blueprint()


from kerko import views  # pylint: disable=wrong-import-position



"""
Functions for extracting data from Zotero items.
"""

import itertools
import re
from abc import ABC, abstractmethod
from collections.abc import Iterable
from datetime import datetime

from flask import current_app
from markupsafe import Markup

from kerko.datetime import maximize_partial_date, parse_partial_date
from kerko.tags import TagGate
from kerko.text import id_normalize, sort_normalize
from kerko.transformers import (find_item_id_in_zotero_uri_links,
                                find_item_id_in_zotero_uris_str)

RECORD_SEPARATOR = '\x1e'


def encode_single(value, spec):
    """Encode a single value."""
    return spec.encode(value)


def encode_multiple(value, spec):
    """Encode items of an iterable."""
    return [spec.encode(item) for item in value]


def is_file_attachment(item, mime_types=None):
    """
    Return `True` if a given item is a file attachment.

    :param mime_types: If a list is provided, the item must also match one of
        the given MIME types. If empty or `None`, the MIME type is not checked.
    """
    if not item.get('data'):
        return False
    if not item['data'].get('key'):
        return False
    if item['data'].get('linkMode') not in ['imported_file', 'imported_url']:
        return False
    if mime_types and item['data'].get('contentType', 'octet-stream') not in mime_types:
        return False
    return True


def is_link_attachment(item):
    """
    Return `True` if a given item is a link attachment.
    """
    if not item.get('data'):
        return False
    if not item['data'].get('key'):
        return False
    if item['data'].get('linkMode') != 'linked_url':
        return False
    if not item['data'].get('url'):
        return False
    return True


class Extractor(ABC):
    """
    Data extractor.

    An extractor can retrieve elements from item or `LibraryContext`` objects,
    and add elements to a document. The document is represented by a `dict`. A
    ``BaseFieldSpec`` object provides both an ``encode()`` method that may
    transform the data before its assignment into the document, and the key to
    assign the resulting data to.
    """

    def __init__(self, format_='data', encode=encode_single, **kwargs):
        """
        Initialize the extractor.

        :param str format_: Format to retrieve when performing the Zotero item
            read, e.g., 'data', 'bib', 'ris', to ensure that the data required
            by this extractor is requested from Zotero to become available in
            the item at extraction time.

        :param callable encode: Function that can encode a value using a
            ``FieldSpec``.
        """
        self.format = format_
        self.encode = encode
        assert not kwargs  # Subclasses should have consumed every keyword arg.

    @abstractmethod
    def extract(self, item, library_context, spec):
        """
        Retrieve the value from context.

        :return: Extracted value, or `None` if no value could be extracted.
        """

    def extract_and_store(self, document, item, library_context, spec):
        """
        Extract value from context and store its encoded version in document.
        """
        extracted_value = self.extract(item, library_context, spec)
        if extracted_value is not None:
            document[spec.key] = self.encode(extracted_value, spec)

    def warning(self, message, item):
        current_app.logger.warning(
            f"{self.__class__.__name__}: {message} ({item.get('key')})"
        )


class TransformerExtractor(Extractor):
    """
    Wrap an extractor to transform data before encoding it into the document.
    """

    def __init__(self, *, extractor, transformers, **kwargs):
        """
        Initialize the extractor.

        :param Extractor extractor: Base extractor to wrap.

        :param list transformers: List of callables that will be chained to
            transform the extracted data. Each callable takes a value as
            argument and returns the transformed value.
        """
        super().__init__(format_=extractor.format, **kwargs)
        self.extractor = extractor
        self.transformers = transformers

    def apply_transformers(self, value):
        for transformer in self.transformers:
            value = transformer(value)
        return value

    def extract(self, item, library_context, spec):
        value = self.extractor.extract(item, library_context, spec)
        return self.apply_transformers(value)


class MultiExtractor(Extractor):
    """
    Allow a composition of multiple extractors.
    """

    def __init__(self, *, extractors, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)
        self.extractors = extractors

    def extract(self, item, library_context, spec):
        values = []
        for extractor in self.extractors:
            assert self.format == extractor.format  # Extractors can only use same format as parent.
            value = extractor.extract(item, library_context, spec)
            if isinstance(value, Iterable) and not isinstance(value, str):
                values.extend(value)
            elif value:
                values.append(value)
        return values or None


class ChainExtractor(Extractor):
    """
    Extract data using a chain of extractors.

    When the an extractor returns `None`, the following one in the chain is
    tried, until a value is found or no more extractors are left to try in the
    chain.
    """

    def __init__(self, *, extractors, **kwargs):
        super().__init__(**kwargs)
        self.extractors = extractors

    def extract(self, item, library_context, spec):
        value = None
        for extractor in self.extractors:
            assert self.format == extractor.format  # Extractors can only use same format as parent.
            value = extractor.extract(item, library_context, spec)
            if value is not None:
                break
        return value


class KeyExtractor(Extractor):  # pylint: disable=abstract-method

    def __init__(self, *, key, **kwargs):
        """
        Initialize the extractor.

        :param str key: Key of the element to extract from the Zotero item.
        """
        super().__init__(**kwargs)
        self.key = key


class ItemExtractor(KeyExtractor):
    """Extract a value from an item."""

    def extract(self, item, library_context, spec):
        return item.get(self.key)


class ItemDataExtractor(KeyExtractor):
    """Extract a value from item data."""

    def extract(self, item, library_context, spec):
        return item.get('data', {}).get(self.key)


class RawDataExtractor(Extractor):

    def extract(self, item, library_context, spec):
        return item.get('data')


class ItemRelationsExtractor(Extractor):
    """Extract a list of item's relations corresponding to a given predicate."""

    def __init__(self, predicate, **kwargs):
        super().__init__(**kwargs)
        self.predicate = predicate

    def extract(self, item, library_context, spec):
        relations = item.get('data', {}).get('relations', {}).get(self.predicate, [])
        if relations and isinstance(relations, str):
            relations = [relations]
        assert isinstance(relations, Iterable)
        return relations


class ItemTypeLabelExtractor(Extractor):
    """Extract the label of the item's type."""

    def extract(self, item, library_context, spec):
        item_type = item.get('data', {}).get('itemType')
        if item_type and item_type in library_context.item_types:
            return library_context.item_types[item_type]
        if item_type != 'attachment':  # Attachment has no label, no need to warn.
            self.warning(f"Missing or unknown item type '{item_type}'", item)
        return None


class ItemFieldsExtractor(Extractor):
    """Extract field metadata."""

    def extract(self, item, library_context, spec):
        item_type = item.get('data', {}).get('itemType')
        if item_type and item_type in library_context.item_fields:
            fields = library_context.item_fields[item_type]
            # Retain metadata for fields that are actually present in the item.
            item_fields = [f for f in fields if f.get('field') in item['data']]
            return item_fields
        if item_type != 'attachment':  # Attachment has no field metadata, no need to warn.
            self.warning(f"Missing or unknown item type '{item_type}'", item)
        return None


class ItemLinkExtractor(Extractor):
    """Extract an item link from the 'links' element."""

    def __init__(self, *, link_key, link_type, **kwargs):
        super().__init__(**kwargs)
        self.link_key = link_key
        self.link_type = link_type

    def extract(self, item, library_context, spec):
        link = item.get('links', {}).get(self.link_key, {})
        if link and link.get('type') == self.link_type:
            return link.get('href')
        return None


class ZoteroWebItemURLExtractor(ItemLinkExtractor):
    """Extract an item's zotero.org link."""

    def __init__(self, *args, **kwargs):
        super().__init__(link_key='alternate', link_type='text/html', *args, **kwargs)


class ZoteroAppItemURLExtractor(Extractor):
    """Extract a link for opening the item in the Zotero app."""

    def extract(self, item, library_context, spec):
        if library_context.library_type == 'group':
            return f"zotero://select/groups/{library_context.library_id}/items/{item.get('key')}"
        return f"zotero://select/library/items/{item.get('key')}"


class CreatorTypesExtractor(Extractor):
    """Extract creator types metadata."""

    def extract(self, item, library_context, spec):
        item_type = item.get('data', {}).get('itemType')
        if item_type and item_type in library_context.creator_types:
            library_creator_types = library_context.creator_types[item_type]
            # Retain metadata for creator types that are actually present in the item.
            item_creator_types = []
            for library_creator_type in library_creator_types:
                for item_creator in item['data'].get('creators', []):
                    creator_type = item_creator.get('creatorType')
                    if creator_type and creator_type == library_creator_type.get('creatorType'):
                        item_creator_types.append(library_creator_type)
                        break
            if item_creator_types:
                return item_creator_types
            if item['data'].get('creators', False):
                self.warning(f"Missing creator types for item type '{item_type}'.", item)
        elif item_type != 'attachment':  # Attachment has no creator types, no need to warn.
            self.warning(f"Missing or unknown item type '{item_type}'", item)
        return None


class CreatorsExtractor(Extractor):
    """Flatten and extract creator data."""

    def extract(self, item, library_context, spec):
        creators = []
        for creator in item.get('data', {}).get('creators', []):
            n = creator.get('name', '').strip()
            if n:
                creators.append(n)
            firstname = creator.get('firstName', '').strip()
            lastname = creator.get('lastName', '').strip()
            if firstname and lastname:
                # Combine firstname and lastname in different orders to help
                # phrase searches.
                creators.append(' '.join([firstname, lastname]))
                creators.append(', '.join([lastname, firstname]))
            elif firstname:
                creators.append(firstname)
            elif lastname:
                creators.append(lastname)
        return RECORD_SEPARATOR.join(creators) if creators else None


class CollectionNamesExtractor(Extractor):
    """Extract item collections for text search."""

    def extract(self, item, library_context, spec):
        names = set()
        for k in item.get('data', {}).get('collections', []):
            if k in library_context.collections:
                name = library_context.collections[k].get('data', {}).get('name', '').strip()
                if name:
                    names.add(name)
        return RECORD_SEPARATOR.join(names) if names else None


class BaseTagsExtractor(Extractor):

    def __init__(self, *, include_re='', exclude_re='', **kwargs):
        """
        Initialize the extractor.

        :param str include_re: Any tag that does not matches this regular
            expression will be ignored by the extractor. If empty, all tags will
            be accepted unless `exclude_re` is set and they match it.

        :param str exclude_re: Any tag that matches this regular expression
            will be ignored by the extractor. If empty, all tags will be
            accepted unless `include_re` is set and they do not match it.
        """
        super().__init__(**kwargs)
        self.include = re.compile(include_re) if include_re else None
        self.exclude = re.compile(exclude_re) if exclude_re else None

    def extract(self, item, library_context, spec):
        tags = set()
        for tag_data in item.get('data', {}).get('tags', []):
            tag = tag_data.get('tag', '').strip()
            if tag and \
                    (not self.include or self.include.match(tag)) and \
                    (not self.exclude or not self.exclude.match(tag)):
                tags.add(tag)
        return tags or None


class TagsTextExtractor(BaseTagsExtractor):
    """Extract item tags for text search."""

    def extract(self, item, library_context, spec):
        tags = super().extract(item, library_context, spec)
        return RECORD_SEPARATOR.join(tags) if tags else None


class BaseChildrenExtractor(Extractor):

    def __init__(self, *, item_type, include_re='', exclude_re='', **kwargs):
        """
        Initialize the extractor.

        :param str item_type: The type of child items to extract, either 'note'
            or 'attachment'.

        :param [str,list] include_re: Any child which does not have a tag that
            matches this regular expression will be ignored by the extractor. If
            empty, all children will be accepted unless `exclude_re` is set and
            causes some to be rejected. When passing a list, every pattern of
            the list must match at least a tag for the child to be included.

        :param [str,list] exclude_re: Any child that have a tag that matches
            this regular expression will be ignored by the extractor. If empty,
            all children will be accepted unless `include_re` is set and causes
            some to be rejected. When passing a list, every pattern of the list
            must match at least a tag for the child to be excluded.
        """
        super().__init__(**kwargs)
        self.item_type = item_type
        self.gate = TagGate(include_re, exclude_re)

    def extract(self, item, library_context, spec):
        accepted_children = []
        for child in item.get('children', []):
            if child.get('data', {}).get('itemType') == self.item_type \
                    and self.gate.check(child.get('data', {})):
                accepted_children.append(child)
        return accepted_children or None


class BaseChildAttachmentsExtractor(BaseChildrenExtractor):

    def __init__(self, **kwargs):
        super().__init__(item_type='attachment', **kwargs)


class ChildFileAttachmentsExtractor(BaseChildAttachmentsExtractor):
    """
    Extract the metadata of stored copies of files into a list of dicts.
    """

    def __init__(self, *, mime_types=None, **kwargs):
        super().__init__(**kwargs)
        self.mime_types = mime_types

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        return [
            {
                'id': child['key'],
                'data': {
                    'contentType': child['data'].get('contentType'),
                    'filename': child['data'].get('filename'),
                    'md5': child['data'].get('md5'),
                    'mtime': child['data'].get('mtime'),
                }
            } for child in children if is_file_attachment(child, self.mime_types)
        ] if children else None


class ChildLinkedURIAttachmentsExtractor(BaseChildAttachmentsExtractor):
    """
    Extract attached links to URIs into a list of dicts.
    """

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return [
                {
                    'title': child['data'].get('title', child['data'].get('url')),
                    'url': child['data'].get('url')
                }
                for child in children if is_link_attachment(child)
            ]
        return None


class ChildAttachmentsFulltextExtractor(BaseChildAttachmentsExtractor):
    """Extract the text content of attachments."""

    def __init__(self, *, mime_types=None, **kwargs):
        super().__init__(**kwargs)
        self.mime_types = mime_types

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return RECORD_SEPARATOR.join(
                [
                    Markup(child['fulltext']).striptags() for child in children
                    if is_file_attachment(child, self.mime_types) and child.get('fulltext')
                ]
            )
        return None


class BaseChildNotesExtractor(BaseChildrenExtractor):  # pylint: disable=abstract-method

    def __init__(self, **kwargs):
        super().__init__(item_type='note', **kwargs)


class ChildNotesTextExtractor(BaseChildNotesExtractor):
    """Extract notes for text search."""

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return RECORD_SEPARATOR.join(
                [
                    Markup(child['data']['note']).striptags()
                    for child in children if child.get('data', {}).get('note')
                ]
            )
        return None


class RawChildNotesExtractor(BaseChildNotesExtractor):
    """Extract raw notes for storage."""

    def extract(self, item, library_context, spec):
        children = super().extract(item, library_context, spec)
        if children:
            return [
                child['data']['note']
                for child in children if child.get('data', {}).get('note')
            ]
        return None


class RelationsInChildNotesExtractor(BaseChildNotesExtractor):
    """Extract item references specified in child notes."""

    def extract(self, item, library_context, spec):
        refs = set()
        children = super().extract(item, library_context, spec)
        if children:
            for child in children:
                note = child.get('data', {}).get('note', '')
                # Find in the href attribute of <a> elements.
                refs.update(find_item_id_in_zotero_uri_links(note))
                # Find in plain text.
                note = Markup(re.sub(r'<br\s*/>', '\n', note)).striptags()
                refs.update(find_item_id_in_zotero_uris_str(note))
        return list(refs) or None


def _expand_paths(path):
    """
    Extract the paths of each of the components of the specified path.

    If the given path is ['a', 'b', 'c'], the returned list of paths is:
    [['a'], ['a', 'b'], ['a', 'b', 'c']]
    """
    return [path[0:i + 1] for i in range(len(path))]


class CollectionFacetTreeExtractor(Extractor):
    """Index the Zotero item's collections needed for the specified facet."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)

    def extract(self, item, library_context, spec):
        # Sets prevent duplication when multiple collections share common ancestors.
        encoded_ancestors = set()
        for collection_key in item.get('data', {}).get('collections', []):
            if collection_key not in library_context.collections:
                continue  # Skip unknown collection.
            ancestors = library_context.collections.ancestors(collection_key)
            if len(ancestors) <= 1 or ancestors[0] != spec.collection_key:
                continue  # Skip collection, unrelated to this facet.

            ancestors = ancestors[1:]  # Facet values come from subcollections.
            for path in _expand_paths(ancestors):
                label = library_context.collections.get(
                    path[-1], {}
                ).get('data', {}).get('name', '').strip()
                encoded_ancestors.add((tuple(path), label))  # Cast path to make it hashable.
        return encoded_ancestors or None


class InCollectionExtractor(Extractor):
    """Extract the boolean membership of an item into a collection."""

    def __init__(self, *, collection_key, true_only=True, check_subcollections=True, **kwargs):
        """
        Initialize the extractor.

        :param str collection_key: Key of the collection to test item membership
            against.

        :param bool true_only: If `True` (default), extraction returns `True`
            when an item belongs to the specified collection, or `None` when it
            does not belong to that collection. If `False`, always return a
            boolean.

        :param bool check_subcollections: If `True` (default), membership is
            extended to any subcollection of the specified collection.
        """
        super().__init__(**kwargs)
        self.collection_key = collection_key
        self.true_only = true_only
        self.check_subcollections = check_subcollections

    def extract(self, item, library_context, spec):
        item_collections = list(
            itertools.chain(
                *[
                    library_context.collections.ancestors(c) if self.check_subcollections else c
                    for c in item.get('data', {}).get('collections', [])
                ]
            )
        )
        is_in = self.collection_key in item_collections
        if not self.true_only:
            return is_in
        if is_in:
            return True
        return None


class TagsFacetExtractor(BaseTagsExtractor):
    """Index the Zotero item's tags for faceting."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)


class ItemTypeFacetExtractor(Extractor):
    """Index the Zotero item's type for faceting."""

    def extract(self, item, library_context, spec):
        item_type = item.get('data', {}).get('itemType')
        if item_type:
            return (item_type, library_context.item_types.get(item_type, item_type))
        self.warning("Missing itemType", item)
        return None


class YearExtractor(Extractor):
    """Parse the Zotero item's publication date to get just the year."""

    def extract(self, item, library_context, spec):
        parsed_date = item.get('meta', {}).get('parsedDate', '')
        if parsed_date:
            year, _month, _day = parse_partial_date(parsed_date)
            return str(year)
        return None


class YearFacetExtractor(Extractor):
    """Index the Zotero item's publication date for faceting by year."""

    def __init__(self, encode=encode_multiple, **kwargs):
        super().__init__(encode=encode, **kwargs)

    def extract(self, item, library_context, spec):
        parsed_date = item.get('meta', {}).get('parsedDate', '')
        if parsed_date:
            year, _month, _day = parse_partial_date(parsed_date)
            decade = int(int(year) / 10) * 10
            century = int(int(year) / 100) * 100
            return _expand_paths([str(century), str(decade), str(year)])
        return None


class ItemDataLinkFacetExtractor(ItemDataExtractor):

    def extract(self, item, library_context, spec):
        return item.get('data', {}).get(self.key, '').strip() != ''


class MaximizeParsedDateExtractor(Extractor):
    """Extract and "maximize" a `datetime` object from the item's `parsedDate` meta field."""

    def extract(self, item, library_context, spec):
        parsed_date = item.get('meta', {}).get('parsedDate', None)
        if parsed_date:
            try:
                return datetime(*maximize_partial_date(*parse_partial_date(parsed_date)))
            except ValueError:
                pass
        return None


def _prepare_sort_text(text):
    """
    Normalize the given text for a sort field.

    :param str text: The Unicode string to normalize.

    :return bytearray: The normalized text.
    """
    return sort_normalize(Markup(text).striptags())


class SortItemDataExtractor(ItemDataExtractor):

    def extract(self, item, library_context, spec):
        return _prepare_sort_text(item.get('data', {}).get(self.key, ''))


class SortCreatorExtractor(Extractor):

    def extract(self, item, library_context, spec):
        creators = []

        def append_creator(creator):
            creator_parts = [
                _prepare_sort_text(creator.get('lastName', '')),
                _prepare_sort_text(creator.get('firstName', '')),
                _prepare_sort_text(creator.get('name', ''))]
            creators.append(' zzz '.join([p for p in creator_parts if p]))

        # We treat creator types like an ordered list, where the first creator
        # type is for primary creators. Depending on the citation style, lesser
        # creator types may not appear in citations. Therefore, we try to sort
        # only by primary creators in order to avoid sorting with data that may
        # be invisible to the user. Only when an item has no primary creator do
        # we fallback to lesser creators.
        for creator_type in library_context.get_creator_types(item.get('data', {})):
            for creator in item.get('data', {}).get('creators', []):
                if creator.get('creatorType', '') == creator_type.get('creatorType'):
                    append_creator(creator)
            if creators:
                break  # No need to include lesser creator types.
        return ' zzzzzz '.join(creators)


class SortDateExtractor(Extractor):

    def extract(self, item, library_context, spec):
        parsed_date = item.get('meta', {}).get('parsedDate', '')
        year, month, day = parse_partial_date(parsed_date)
        return int('{:04d}{:02d}{:02d}'.format(year, month, day))



from flask import url_for

from kerko.shortcuts import composer


def format_creator_name(creator):
    """Format a name from a dict having either a 'name' key, or 'firstName' and 'lastName' keys."""
    return creator.get(
        'name',
        ', '.join(
            [
                name_part for name_part in [
                    creator.get('lastName', '').strip(),
                    creator.get('firstName', '').strip()
                ] if name_part
            ]
        ),
    )


def inject_creator_display_names(item, link=True):
    """
    Inject the display names of creators into the given item.

    :param dict item: The item whose creators are to be updated.

    :param bool link: Whether to provide a search link for the creator.
    """
    for spec in item.keys():
        if spec == 'data' and 'creators' in item['data']:
            for creator in item['data']['creators']:
                # Add creator display name.
                creator['display'] = format_creator_name(creator)
                # Add creator type labels.
                if 'creator_types' in item:
                    for t in item['creator_types']:
                        if t['creatorType'] == creator['creatorType']:
                            creator['label'] = t['localized']
                            break
                if link:
                    creator_scope = composer().scopes.get('creator')
                    if creator_scope:
                        creator['url'] = url_for(
                            '.search',
                            **creator_scope.add_keywords(value=f"\"{creator['display']}\""),
                        )



"""Item view building functions."""

from urllib.parse import urlparse

from flask import request, url_for

from kerko.shortcuts import config
from kerko.views.item import creators, facets, meta, relations


def is_valid_url(url):
    parsed_url = urlparse(url)
    return parsed_url.scheme and parsed_url.netloc


def inject_item_data(item):
    creators.inject_creator_display_names(item)
    relations.inject_relations(item)
    facets.inject_facet_results(item)


def build_item_context(item):
    context = {
        'item':
            item,
        'item_url':
            url_for('.item_view', item_id=item['id'], _external=True),
        'title':
            item.get('data', {}).get('title', ''),
        'highwirepress_tags':
            meta.build_highwirepress_tags(item),
    }
    if config('kerko.features.open_in_zotero_app'):
        context['open_in_zotero_app'] = bool(request.cookies.get('open-in-zotero-app'))
        url = item.get('zotero_app_url', '')
        if url and is_valid_url(url):
            context['open_in_zotero_app_url'] = url
    if config('kerko.features.open_in_zotero_web'):
        context['open_in_zotero_web'] = bool(request.cookies.get('open-in-zotero-web'))
        url = item.get('zotero_web_url', '')
        if url and is_valid_url(url):
            context['open_in_zotero_web_url'] = url
    return context



"""Utilities for preparing embedded metadata for HTML pages."""

import re

from flask import url_for

from kerko.shortcuts import config
from kerko.views.item.creators import format_creator_name


def build_highwirepress_tags(item):  # pylint: disable=too-many-branches
    # References:
    # - https://scholar.google.ca/intl/en/scholar/inclusion.html#indexing
    # - https://www.monperrus.net/martin/accurate+bibliographic+metadata+and+google+scholar
    tags = []
    data = item.get('data', {})
    if config('kerko.meta.highwirepress_tags') and data.get('itemType') in [
        'book',
        'conferencePaper',
        'journalArticle',
        'report',
        'thesis',
    ]:
        tags.append(('citation_title', data.get('title', data.get('shortTitle', '')).strip()))
        tags.append(('citation_publication_date', data.get('date', '')))
        tags.append(('citation_date', data.get('date', '')))
        tags.append(('citation_year', item.get('year', '')))

        for creator in data.get('creators', []):
            if creator.get('creatorType', '') == 'author':
                # Google Scholar requires that only the actual authors be included.
                tags.append(('citation_author', format_creator_name(creator)))

        for tag in ['volume', 'issue', 'language', 'publisher']:
            if data.get(tag):
                tags.append((f'citation_{tag.lower()}', data[tag]))

        pages = data.get('pages')
        if pages:
            pages = re.sub('[-\u2013]', '-', pages)  # Replace hyphens and en-dashes.
            pages = [p.strip() for p in re.split(r'\W*-\W*', data['pages'], 2)]
            if pages and pages[0]:
                tags.append(('citation_firstpage', pages[0]))
            if len(pages) > 1 and pages[1]:
                tags.append(('citation_lastpage', pages[1]))

        if data['itemType'] == 'conferencePaper':
            conference_title = data.get('conferenceName', '').strip()
            if conference_title:
                tags.append(('citation_conference_title', conference_title))
        elif data['itemType'] == 'journalArticle':
            publication_title = data.get(
                'publicationTitle',
                data.get('journalAbbreviation', ''),
            ).strip()
            if publication_title:
                tags.append(('citation_journal_title', publication_title))
        elif data['itemType'] == 'thesis':
            university = data.get('university', '').strip()
            if university:
                tags.append(('citation_dissertation_institution', university))
        elif data['itemType'] == 'report':
            institution = data.get('institution', '').strip()
            if institution:
                tags.append(('citation_technical_report_institution', institution))
            report_number = data.get('reportNumber', '').strip()
            if report_number:
                tags.append(('citation_technical_report_number', report_number))

        for tag in ['ISBN', 'ISSN', 'DOI']:
            if data.get(tag):
                tags.append((f'citation_{tag.lower()}', data[tag]))
            elif data.get('extra'):  # Look for it in the Extra field.
                matches = re.search(
                    fr'^\s*({tag}):\s*(\S+)\s*$', data['extra'], re.IGNORECASE | re.MULTILINE
                )
                if matches:
                    tags.append((f'citation_{tag.lower()}', matches[2]))

        for attachment in item.get('attachments', []):
            if attachment.get('data', {}).get('contentType') == 'application/pdf':
                tags.append(
                    (
                        'citation_pdf_url',
                        url_for(
                            '.child_attachment_download',
                            item_id=item['id'],
                            attachment_id=attachment['id'],
                            attachment_filename=attachment['data'].get('filename', item['id']),
                            _external=True
                        )
                    )
                )
                break  # Only attach the first PDF found.

    return tags
